#include "myqtapp.h"

#include "fit_W.cpp" // TODO: Not so nice...



// Because itoa is not supported by gcc...
/**
 * C++ version 0.4 char* style "itoa":
 * Written by Lukás Chmela
 * Released under GPLv3.
 */
char* itoa(int value, char* result, int base) {
	// check that the base if valid
	if (base < 2 || base > 36) { *result = '\0'; return result; }

	char* ptr = result, *ptr1 = result, tmp_char;
	int tmp_value;

	do {
		tmp_value = value;
		value /= base;
		*ptr++ = "zyxwvutsrqponmlkjihgfedcba9876543210123456789abcdefghijklmnopqrstuvwxyz" [35 + (tmp_value - value * base)];
	} while ( value );

	// Apply negative sign
	if (tmp_value < 0) *ptr++ = '-';
	*ptr-- = '\0';
	while(ptr1 < ptr) {
		tmp_char = *ptr;
		*ptr--= *ptr1;
		*ptr1++ = tmp_char;
	}
	return result;
}



// To manage the files within a directory.
struct image_file {

	string file_name;
	
	// A number to store the random position of this file in the
	// order. (Verblindung)
	int random_pos;
};
bool operator< (const image_file& a, const image_file& b) {
  return (a.random_pos < b.random_pos);
}



// Global variables:

// All files within the current directory.
vector<image_file> all_jpg;
// Storing the position in all_jpg of the file currently looked at.
unsigned int all_jpg_pos;
// Working directory path.
string dir_path;
// The region of interest in the file shall remain over all files.
int roi_x = 0;
int roi_y = 0;
int roi_width = 0;
int roi_height = 0;

// The top and bottom cut-off values.
int cutoff_top = 1;
int cutoff_bottom = 5;

// Current path.
QString path = "C:/";

myQtApp::myQtApp(QWidget *parent)
{
	setupUi( this ); // this sets up GUI

	// signals/slots mechanism in action
	connect( pushButton_browse, SIGNAL( clicked() ), this, SLOT( get_path() ) );  
	connect( pushButton_process, SIGNAL( clicked() ), this, SLOT( process_directory() ) ); 
	connect( pushButton_file_fwd, SIGNAL( clicked() ), this, SLOT( file_fwd() ) );
	connect( pushButton_file_bwd, SIGNAL( clicked() ), this, SLOT( file_bwd() ) );
	connect( pushButton_do, SIGNAL( clicked() ), this, SLOT( do_something() ) );
}

// Returns the number of measures stored in measures.csv for a specific file.
// (Not efficient, but I think in our case it does not matter because there
// will never be more than 100 files in a directory...)
int myQtApp::get_number_of_measures(string file_name) {

	ifstream file((dir_path + "/measures.csv").c_str());
	// If no file exists return 0.
	if (!file) {
		return 0;
	}

	int counter = 0;
	string line;
	while(getline(file, line)) {
		// Find first semicolon.
		unsigned int i; 
		for (i=0; i<line.length(); i++) {
			if (line.substr(i,1).compare(";") == 0) {
				break;
			}
		}
		if (line.substr(0,i).compare(file_name) == 0) {
			counter++;
		}
	}
		
	file.close();
	return counter;
}


void myQtApp::update_file_list() {
	// Write all files into the text field.
	textBrowser_file_overview->clear();
	
	string tmp_str = "";
	for (unsigned int i=0; i<all_jpg.size(); i++) {
	
		int num_measures = get_number_of_measures(all_jpg[i].file_name);
				
		char buffer_pos[16];
		itoa(i+1, buffer_pos, 10);
		char buffer_m[16];
		itoa(num_measures, buffer_m, 10);
		
		// Set the scrollTo-Anchor three lines before the current file.
		if (i == all_jpg_pos-3) {
			tmp_str += "<a name =\"hier\"></a>";
		}
		// Highlight the current file. &#60; is escape character for <
		if (i == all_jpg_pos) {
			tmp_str += "<b>File " + string(buffer_pos) + " (" + string(buffer_m) + " measures) &#60;-- </b><br>";
		}
		else {
			tmp_str += "File " + string(buffer_pos) + " (" + string(buffer_m) + " measures)<br>";
		}
	}
	
	textBrowser_file_overview->setText(tmp_str.c_str());
	textBrowser_file_overview->scrollToAnchor("hier");
}
	
void myQtApp::file_fwd() {

	if (all_jpg.size() == 0) {
		cout << "No jpg containing directory selected." << endl;
		return;
	}

	// If the user entered something in the jump to field.
	if (lineEdit_jump_to->isModified()) {
		unsigned int jump_to = atoi(lineEdit_jump_to->text().toLatin1().data());
		
		// Catch invalid entries.
		if (jump_to > all_jpg.size()) {
			jump_to = all_jpg.size();
		}
		
		all_jpg_pos = jump_to - 1;
		lineEdit_jump_to->setText("");
	}

	else {
		if (all_jpg_pos < all_jpg.size()-1) {
			++all_jpg_pos;
		}
		else {
			all_jpg_pos = 0;
		}
	}

	update_file_list();
}

void myQtApp::file_bwd() {

	if (all_jpg.size() == 0) {
		cout << "No jpg containing directory selected." << endl;
		return;
	}

	// If the user entered something in the jump to field.
	if (lineEdit_jump_to->isModified()) {
		unsigned int jump_to = atoi(lineEdit_jump_to->text().toLatin1().data());
		
		// Catch invalid entries.
		if (jump_to > all_jpg.size()) {
			jump_to = all_jpg.size();
		}
		
		all_jpg_pos = jump_to - 1;
		lineEdit_jump_to->setText("");
	}

	else {
		if (all_jpg_pos > 0) {
			--all_jpg_pos;
		}
		else {
			all_jpg_pos = all_jpg.size()-1;
		}
	}
		
	update_file_list();
}

void myQtApp::get_path()
{
	// Get path from text field.
	path = QFileDialog::getExistingDirectory(
		this,
		"Choose a file to open",
		path);
		
	lineEdit_directory->setText(path);
}


void myQtApp::process_directory() {

	// Put the path into a normal string.
	dir_path = lineEdit_directory->text().toLatin1().data();
	// In case no entry has been made return.
	if (dir_path.length() < 1) return;

	DIR* dip;
    struct dirent* dit;
	
	if ((dip = opendir(dir_path.c_str())) == NULL) {
		cout << dir_path << " is no valid directory." << endl;
		return;
    }
	
	// Check if there already exists a random order for the files in
	// this directory. (Stored in the random_order.csv file.)
	bool has_order = false;
	
	// Clear the global vector for all image files in the directory.
	all_jpg.clear();
	all_jpg_pos = 0;
	
	image_file next_entry;
	while ((dit = readdir(dip)) != NULL) {		
		// We are only interested in .jpg files.
		string file_name = dit->d_name;
		//cout << file_name << endl;
		if (file_name.length() > 5) {
			if (( file_name.substr(file_name.length()-4, 4).compare(".jpg") == 0 ) ||
				( file_name.substr(file_name.length()-4, 4).compare(".JPG") == 0 )) {
			
				//cout << "accepted" << endl;
				next_entry.file_name = file_name;
				all_jpg.push_back(next_entry);
			}
			else if (file_name.compare("random_order.csv") == 0) {
				has_order = true;			
			}
		}
	}
	
	selected_dir->setText(QString(dir_path.c_str()));
	
	if (all_jpg.size() == 0) {
		cout << "No jpg containing directory selected." << endl;
		textBrowser_file_overview->setText(QString("Please select a jpg containing directory."));
		return;
	}
  
	// Closing directory.
    if (closedir(dip) == -1) {
        perror("closedir");
    }
    
	// If no random order exists yet, create one.
	if (!has_order) {
	
		ofstream file((dir_path + "/random_order.csv").c_str());
		//cout << "Creating new order: " << (dir_path + "/random_order.csv") << endl;

		srand(time(NULL));
		
		for (unsigned int i=0; i<all_jpg.size(); i++) {
			all_jpg[i].random_pos = rand();
			file << all_jpg[i].file_name << ";";
			file << all_jpg[i].random_pos << ";\n";
			//cout << i << " : " << all_jpg[i].file_name << ";" << all_jpg[i].random_pos << endl;
		}
		
		file.close();
	}
		
	// If there is a random order already, take its values.
	else {
	
		ifstream file((dir_path + "/random_order.csv").c_str());
  		if (!file) {
			cout << "Unable to open file " << dir_path + "/random_order.csv" << endl;
			return;
		}
	
		//cout << "Using existing order: " << (dir_path + "/random_order.csv") << endl;
	
		// Recreate the all_jpg vector from random_pos entries.
		all_jpg.clear();
		all_jpg_pos = 0;
		image_file next_entry;
		string line;
		int last_sem = 0;
		while(getline(file, line)) {
			last_sem = 0;
			for (unsigned int i=0; i<line.length(); i++) {
				if (line.substr(i,1).compare(";") == 0) {
					if (last_sem == 0) {
						//cout << line.substr(0,i) << endl;
						next_entry.file_name = line.substr(0,i);
						last_sem = i;
					}
					else {
						//cout << line.substr(last_sem+1, i-last_sem-1) << endl << endl;
						next_entry.random_pos = atoi(line.substr(last_sem+1, i-last_sem-1).c_str());
						all_jpg.push_back(next_entry);
						break;
					}
				}
			}
		}
		
		file.close();
	}
	
	std::sort(all_jpg.begin(), all_jpg.end());

	update_file_list();

}


// Disable all buttons while the openCV windows are open.
void myQtApp::lock_all() {

	string tmp_str = "<span style=\"font-family:\'Fixedsys\'; font-size:8pt;\">";
	
	tmp_str += "Please use the following keys on the OpenCV window:<br><br>";

	tmp_str += "ENTER: Save measurement<br>";
	tmp_str += "ESC: Cancel measurement<br><br>";

	tmp_str += "m: Set w-curve top and bottom by mouse.<br><br>";

	tmp_str += "s: Select region of interest by mouse.<br>";
	tmp_str += "q: Reset region of interest.<br><br>";

	tmp_str += "o: Optimiser (all values)<br>";
	tmp_str += "p: Optimiser (width->lambda->all)<br>";
	tmp_str += "u: Optimiser (width)<br>";
	tmp_str += "i: Optimiser (lambda)<br><br>";

	tmp_str += "w: Reposition (width)<br>";
	tmp_str += "h: Reposition (height)<br>";
	tmp_str += "t: Reposition (translation)<br>";
	tmp_str += "l: Reposition (lambda)<br>";
	tmp_str += "r: Reposition (all values)<br><br>";
	
	tmp_str += "d: Show visualisation of distance transformation";
	
	tmp_str += "</span>";

	textBrowser_file_overview->setText(tmp_str.c_str());


	pushButton_do->setEnabled(false);
	pushButton_file_fwd->setEnabled(false);
	pushButton_file_bwd->setEnabled(false);
	pushButton_browse->setEnabled(false);
	pushButton_process->setEnabled(false);
	lineEdit_threshold->setEnabled(false);
	lineEdit_jump_to->setEnabled(false);
	lineEdit_lambda->setEnabled(false);
	lineEdit_directory->setEnabled(false);
}

// Enable buttons again after go_fit_W is finished.
void myQtApp::unlock_all() {
	pushButton_do->setEnabled(true);
	pushButton_file_fwd->setEnabled(true);
	pushButton_file_bwd->setEnabled(true);
	pushButton_browse->setEnabled(true);
	pushButton_process->setEnabled(true);
	lineEdit_threshold->setEnabled(true);
	lineEdit_jump_to->setEnabled(true);
	lineEdit_lambda->setEnabled(true);
	lineEdit_directory->setEnabled(true);
}

void myQtApp::do_something() {

	if (all_jpg.size() == 0) {
		cout << "No jpg containing directory selected." << endl;
		return;
	}

	// The threshold can be given and taken from the processed file.
	int thresh = atoi(lineEdit_threshold->text().toLatin1().data());
	
	// Variables to store the returned measurements.
	double lambda = (double) atoi(lineEdit_lambda->text().toLatin1().data()) / 1000;
	double energy = 0.0;	

	// Lock all buttoms and fields.
	lock_all();
	
	cout << endl << "Starting new measurement for file " << all_jpg_pos+1 << ":" << endl;


	// For debug!
	// thresh = 165;
	// roi_x = 1046;
	// roi_y = 716;
	// roi_width = 1616-1046;
	// roi_height = 1312-716;
	// go_fit_W("C:/repos/bud/fotos/DSCN6411.JPG",
		// thresh,
		// lambda,
		// energy,
		// roi_x,
		// roi_y,
		// roi_width,
		// roi_height,
		// cutoff_top,
		// cutoff_bottom
	// );
	
	// Run go_fit_W for current file.
	go_fit_W((dir_path +"/"+ all_jpg[all_jpg_pos].file_name).c_str(),
		thresh,
		lambda,
		energy,
		roi_x,
		roi_y,
		roi_width,
		roi_height,
		cutoff_top,
		cutoff_bottom
	);
	
	// If ESC was pressed instead of ENTER, do not safe the values.
	if (lambda == 9999) {
		unlock_all();
		cout << "Measurement for file " << all_jpg_pos+1 << " CANCELLED." << endl << endl;
		// Update the file list.
		update_file_list();
		return;
	}
	
	// Check if the measures.csv file already exists.
	bool file_exist = false;
	ifstream ifile((dir_path + "/measures.csv").c_str());
	if (ifile) {
		file_exist = true;
		ifile.close();
	}
	
	
	// Prepare time stamp.
	time_t ltime = time(NULL);
	struct tm *Tm = localtime(&ltime);
	
	
	// Prepare the "date picture taken".
	char buf[1024];
	
	ExifData* ed = exif_data_new_from_file((dir_path +"/"+ all_jpg[all_jpg_pos].file_name).c_str());
    if (!ed) {
        cout << "File not readable or no EXIF data in file " << (dir_path +"/"+ all_jpg[all_jpg_pos].file_name).c_str() << endl;
    }
	else {
	
		ExifIfd ifd = EXIF_IFD_0;
		ExifTag tag = (ExifTag) EXIF_TAG_DATE_TIME;
		ExifEntry* ee = exif_content_get_entry(ed->ifd[ifd], tag);
		// See if this tag exists.
		if (ee) {
			// Get the contents of the tag in human-readable form.
			exif_entry_get_value(ee, buf, sizeof(buf));

			// Change the format from YYYY:MM:DD to YYYY-MM-DD.
			unsigned int i = 0;
			unsigned int j = 0;
			while ((i<sizeof(buf)) && (j<2)) {
				if (buf[i] == ':') {
					j++;
					buf[i] = '-';
				}
				i++;
			}
		}
		else {
			cout << "Exif entry not found!" << endl;
		}
		// Free the exif data.
		exif_data_unref(ed);
	}
		
		
	// Open file in append mode.
	fstream file;
	file.open((dir_path + "/measures.csv").c_str(), ios::out | ios::app);
	if (!file.is_open()) {
		cout << "ERROR opening file" << endl;
		return;
	}
	
	// If file had to be created, place column titles.
	if (!file_exist) {
		file << "Dateiname;Aufnahmedatum;Vermessungsdatum;Lambda;Energie;\n"; 
	}
		
		
	// Write the measures for this file.
	file << all_jpg[all_jpg_pos].file_name << ";"; // file name
	file << buf << ";"; // date picture taken
	file << setfill('0');
	file << Tm->tm_year +1900 << "-" << setw(2) << Tm->tm_mon +1 << "-" << setw(2) << Tm->tm_mday << " "; // time stamp day
	file << setw(2) << Tm->tm_hour << ":" << setw(2) << Tm->tm_min << ":" << setw(2) << Tm->tm_sec << ";"; // time stamp time
	file << lambda << ";" << energy << ";\n";
		
	file.close();
	
	cout << "Measurement for file " << all_jpg_pos+1 << " SAVED." << endl << endl;
	
	// Unlock the buttoms and fields again.
	unlock_all();
		
	// Insert the received threshold value into the GUI field.
	char buffer_pos[16];
	
	itoa(thresh, buffer_pos, 10);
	
	lineEdit_threshold->setText(buffer_pos);
	
	// Update the file list.
	update_file_list();
	
}




